/*
  maildirpasswd.c
  by Tom Clegg <tom@fiction.org>
  copyright 2001
  licensed under GPL
  see file COPYING for details

  $Id: maildirpasswd.c,v 0.6 2006/01/23 21:50:43 tom Exp $
*/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
/* #include <crypt.h> */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define BUF 255

main(int argc, char** argv)
{
  char hostname[BUF], newpass[BUF], newpass2[BUF], passwdfile[BUF];
  char* user = getenv("USER");
  char* home = getenv("HOME");
  char* ext = argv[1];
  int passlen, i, fnlen, fd;
  pid_t child;
  char salt[2];
  const char* crypted;

  if(argc!=2) { fprintf(stderr,"Usage: %s ext\n",argv[0]); exit(1); }
  if(!user || !home)
    { fprintf(stderr,"Sorry, $USER and $HOME must be set\n"); exit(1); }

  gethostname(hostname,BUF);
  if(isatty(0))
    printf("Changing password for POP mailbox %s-%s@%s\nNew password: ",
	   getenv("USER"), ext, hostname);
  fflush(stdout);

  if(isatty(2) && !(child=fork())) execlp("stty","stty","-echo",0);
  waitpid(child,0,0);
  passlen=read(0,newpass,BUF);
  if(passlen<=0) {
    extern int errno;
    fprintf(stderr,"Couldn't read password: %s\n",strerror(errno));
    exit(1);
  }
  newpass[passlen]='\0';
  if(isatty(2)) {
    printf("\nNew password (again): ");
    fflush(stdout);
    if(passlen!=read(0,newpass2,BUF)) newpass2[0]=0;
    if(!(child=fork())) execlp("stty","stty","echo",0);
    waitpid(child,0,0);
    printf("\n");
  } else {
    strcpy(newpass2,newpass);
  }

  if(passlen<=0)
    exit(1);
  newpass[passlen-1]=0;
  newpass2[passlen-1]=0;
  if(strcmp(newpass,newpass2)) {
    fprintf(stderr,"Passwords do not match.  Mailbox password unchanged.\n");
    exit(1);
  }
  srand(time(0));
  for(i=0; i<2; i++) {
    salt[i] = 1+(char) (127.0*rand()/(RAND_MAX+1.0));
    if(!isalnum(salt[i]))
      --i;
  }
  crypted=crypt(newpass,salt);
  if(isatty(0)) {
    printf("Ok, putting encrypted password %s in ~/Maildir-%s/passwd\n",
	   crypted, ext);
  }
  fnlen=snprintf(passwdfile, BUF, "%s/Maildir-%s/passwd", home, ext);
  if(fnlen>BUF || fnlen<=0) {
    fprintf(stderr,"Aak, password filename too long\n");
    exit(1);
  }
  if(0>=(fd=open(passwdfile,O_WRONLY|O_CREAT|O_TRUNC,0x180)))
    { fprintf(stderr,"Can't create password file %s\n", passwdfile); exit(1); }
  if(0!=chmod(passwdfile,0x180))
    { fprintf(stderr,"Can't set permissions on %s\n", passwdfile); exit(1); }
  if(strlen(crypted)!=write(fd,crypted,strlen(crypted)))
    { fprintf(stderr,"Can't write password file %s\n", passwdfile); exit(1); }
  close(fd);
  exit(0);
}
