/*
  maildirpop3d.c
  by Tom Clegg <tom@fiction.org>
  copyright 2001
  licensed under GPL
  see file COPYING for details

  $Id: maildirpop3d.c,v 0.6 2006/01/23 21:50:43 tom Exp $
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/* #include <crypt.h> */
#include "config.h"
#ifdef HAVE_MD5
#include <sys/md5.h>
#endif

#define BUF 256
const char* qmail_pop3d;

static int dovecot_insteadof_exec=0;

void die(char* s)
{
  printf("-ERR %s\n",s);
  exit(1);
}

void nomem()
{
  die("Out of memory");
}

void invchar(const char* inwhat)
{
  printf("-ERR Invalid character in %s\n",inwhat);
  exit(1);
}

checkpassword(const char* Maildir, const char* pass, const char* apop)
{
  int fd, rcsize;
  char realcrypted[BUF];

  if(chdir(Maildir))
    { printf("-ERR Can't chdir to %s\n", Maildir); exit(1); }
  if(0>=(fd=open("passwd",O_RDONLY)))
    { printf("-ERR Can't open %s/passwd\n", Maildir); exit(1); }
  rcsize=read(fd,realcrypted,BUF-1);
  if(rcsize<10)
    { printf("-ERR Can't read %s/passwd\n", Maildir); exit(1); }
  realcrypted[rcsize]=0;
  close(fd);
  if(chdir(".."))
    { printf("-ERR Can't chdir out of %s\n", Maildir); exit(1); }

  if(strcmp(realcrypted,crypt(pass,realcrypted)))
    { printf("-ERR Wrong password\n"); exit(1); }
}

exec_pop3d(const char* Maildir,
	   const char* user,
	   const char* ext,
	   struct passwd* user_passwd,
	   const char* pass,
	   const char* apop)
{
  char* maildir;
  char* homeenv;
  char* userenv;
  FILE* fd4;

  if (chdir (user_passwd->pw_dir)) {
    printf("-ERR Couldn't chdir to %s\n", user_passwd->pw_dir);
    exit(1);
  }
  if (ext) {
    if(!(maildir=malloc(strlen(Maildir)+strlen(ext)+2))) nomem();
    strcpy(maildir,Maildir);
    strcat(maildir,"-");
    strcat(maildir,ext);
    Maildir=maildir;
  }
  setuid(user_passwd->pw_uid);
  checkpassword(Maildir,pass,apop);

  /* set HOME and USER environment vars */

  if(!(homeenv=malloc(strlen(user_passwd->pw_dir)+6))) nomem();
  strcpy(homeenv,"HOME=");
  strcpy(homeenv+5,user_passwd->pw_dir);
  putenv(homeenv);

  if(!(userenv=malloc(strlen(user_passwd->pw_name)+6))) nomem();
  strcpy(userenv,"USER=");
  strcpy(userenv+5,user_passwd->pw_name);
  putenv(userenv);

  if (dovecot_insteadof_exec) {
    if (strchr(user_passwd->pw_name, '\t')) invchar("login name");
    if (strchr(user_passwd->pw_dir, '\t')) invchar("home directory");
    if (strchr(Maildir, '\t')) invchar("maildir");
    fd4 = fdopen(4,"w");
    if (fd4 != NULL &&
	fprintf(fd4,
		"userdb_user=%s\t"
		"userdb_home=%s/%s\t"
		"userdb_uid=%d\t"
		"userdb_gid=%d\t",
		user_passwd->pw_name,
		user_passwd->pw_dir,
		Maildir,
		user_passwd->pw_uid,
		user_passwd->pw_gid)) {
      exit(0);
    }
    else {
      die("Can't write to fd 4");
      exit(111);
    }
  }

  execlp(qmail_pop3d,qmail_pop3d,Maildir,0);
  exit(0);
}

const char* lookup_virtualdomains(const char *domain)
{
  static char buf[1000];
  int buflen;
  int fd;
  size_t readbytes;
  char *newline;
  int skiplongline;
  int colonpos;

  fd = open("/var/qmail/control/virtualdomains", O_RDONLY);
  if (fd) {
    skiplongline = 0;
    buflen = 0;
    readbytes = 1;
    buf[0] = '\0';
    while ((newline = strchr(buf, '\n')) || readbytes > 0) {
      if (newline == 0) {
	if (buflen == sizeof(buf)-1) {
	  buflen = 0;
	  skiplongline = 1;
	}
	readbytes = read(fd, &buf[buflen], sizeof(buf)-1-buflen);
	if (readbytes > 0)
	  buflen += readbytes;
	buf[buflen] = '\0';
	continue;
      }
      if (!skiplongline) {
	colonpos = strlen(domain);
	if (0 == strncmp(buf, domain, colonpos)
	    && buf[colonpos] == ':') {
	  while(newline>buf+colonpos && strchr(" \r\t", newline[-1]))
	    --newline;
	  *newline = '\0';
	  close(fd);
	  return &buf[strlen(domain)+1];
	}
      }
      skiplongline = 0;
      memcpy(buf, newline+1, buflen-(newline-buf));
      buflen -= (newline-buf)+1;
    }
  }
  close(fd);
  return 0;
}

main(int argc, char** argv)
{
  char buf[BUF], user[BUF], pass[BUF], apop[BUF], *ext;
  char *atdomain;
  const char *virtualuser;

  ssize_t sz;
  struct passwd* user_passwd;
  const char* Maildir;
  const char* checkpassword;
  const char* argv0;

  if (argv[0]) {
    argv0 = strrchr(argv[0], '/');
    if (argv0 == 0)
      argv0 = argv[0];
    if (strstr(argv0, "dovecot"))
      dovecot_insteadof_exec=1;
  }

  if(argc>1)
    qmail_pop3d=argv[1];
  else
    qmail_pop3d="/var/qmail/bin/qmail-pop3d";

  if(argc>2)
    Maildir = argv[2];
  else
    Maildir = "Maildir";

  if(argc>3)
    checkpassword = argv[3];
  else
    checkpassword = "checkpassword";

  sz=read(3,buf,BUF-1);
  close(3);
  if(sz<=0) {
    printf("-ERR Can't read from fd 3\n");
    exit(1);
  }
  buf[sz]=0;
  strcpy(user,buf);
  strncpy(pass,buf+strlen(user)+1,sz-strlen(user)-1);
  strncpy(apop,buf+strlen(user)+2+strlen(pass),sz-strlen(user)-2-strlen(pass));

  if ((user_passwd = getpwnam(user))) {
    int pfd[2];
    pid_t child;

    if(pipe(pfd)) die("Can't create pipe");
    if((child=fork())) { /* parent => regular checkpassword qmail-pop3d ... */
      extern int errno;
      close(pfd[1]);
      dup2(pfd[0],3);
      if(checkpassword[0]=='/')
	execl(checkpassword,"checkpassword",qmail_pop3d,Maildir,0);
      else
	execlp(checkpassword,"checkpassword",qmail_pop3d,Maildir,0);
      printf("-ERR Can't exec %s: %s\n", checkpassword, strerror(errno));
      exit(1);
    }
    else { /* child => propagate ``username\0password\0'' on fd 3 */
      close(pfd[0]);
      dup2(pfd[1],3);
      write(3,user,strlen(user)+1);
      write(3,pass,strlen(pass)+1);
      write(3,apop,strlen(apop)+1);
      close(3);
      exit(0);
    }
  }
  else {
    for (atdomain=user; *atdomain; atdomain++) {
      if (*atdomain == '@') {
	*atdomain = '\0';
	if ((virtualuser = lookup_virtualdomains(atdomain+1))
	    && (user_passwd = getpwnam(virtualuser))) {
	  exec_pop3d(Maildir, virtualuser, user, user_passwd, pass, apop);
	}
	*atdomain = '@';
      }
    }
    for (ext=user+strlen(user)-1; ext>user; ext--) {
      if (*ext != '-') {
	if (!isalnum(*ext) && !strchr(":;,._=+~`!#$^&*()[]{}|?'\"",*ext))
	  break;
	continue;
      }
      *ext = '\0';
      if ((user_passwd = getpwnam(user))) {
	exec_pop3d(Maildir, user, ext+1, user_passwd, pass, apop);
      }
      *ext = '-';
    }
  }
  printf("-ERR Couldn't find a user called %s\n", user);
  exit(1);
}
